import re
import sys
from typing import Any

import yaml
from boto3 import Session
from boto3.dynamodb.conditions import Attr
from schema import Optional, Schema

from prowler.lib.logger import logger
from prowler.lib.outputs.utils import unroll_tags

mutelist_schema = Schema(
    {
        "Accounts": {
            str: {
                "Checks": {
                    str: {
                        "Regions": list,
                        "Resources": list,
                        Optional("Tags"): list,
                        Optional("Exceptions"): {
                            Optional("Accounts"): list,
                            Optional("Regions"): list,
                            Optional("Resources"): list,
                            Optional("Tags"): list,
                        },
                    }
                }
            }
        }
    }
)


def parse_mutelist_file(
    mutelist_path: str, aws_session: Session = None, aws_account: str = None
):
    try:
        # Check if file is a S3 URI
        if re.search("^s3://([^/]+)/(.*?([^/]+))$", mutelist_path):
            bucket = mutelist_path.split("/")[2]
            key = ("/").join(mutelist_path.split("/")[3:])
            s3_client = aws_session.client("s3")
            mutelist = yaml.safe_load(
                s3_client.get_object(Bucket=bucket, Key=key)["Body"]
            )["Mutelist"]
        # Check if file is a Lambda Function ARN
        elif re.search(r"^arn:(\w+):lambda:", mutelist_path):
            lambda_region = mutelist_path.split(":")[3]
            lambda_client = aws_session.client("lambda", region_name=lambda_region)
            lambda_response = lambda_client.invoke(
                FunctionName=mutelist_path, InvocationType="RequestResponse"
            )
            lambda_payload = lambda_response["Payload"].read()
            mutelist = yaml.safe_load(lambda_payload)["Mutelist"]
        # Check if file is a DynamoDB ARN
        elif re.search(
            r"^arn:aws(-cn|-us-gov)?:dynamodb:[a-z]{2}-[a-z-]+-[1-9]{1}:[0-9]{12}:table\/[a-zA-Z0-9._-]+$",
            mutelist_path,
        ):
            mutelist = {"Accounts": {}}
            table_region = mutelist_path.split(":")[3]
            dynamodb_resource = aws_session.resource(
                "dynamodb", region_name=table_region
            )
            dynamo_table = dynamodb_resource.Table(mutelist_path.split("/")[1])
            response = dynamo_table.scan(
                FilterExpression=Attr("Accounts").is_in([aws_account, "*"])
            )
            dynamodb_items = response["Items"]
            # Paginate through all results
            while "LastEvaluatedKey" in dynamodb_items:
                response = dynamo_table.scan(
                    ExclusiveStartKey=response["LastEvaluatedKey"],
                    FilterExpression=Attr("Accounts").is_in([aws_account, "*"]),
                )
                dynamodb_items.update(response["Items"])
            for item in dynamodb_items:
                # Create mutelist for every item
                mutelist["Accounts"][item["Accounts"]] = {
                    "Checks": {
                        item["Checks"]: {
                            "Regions": item["Regions"],
                            "Resources": item["Resources"],
                        }
                    }
                }
                if "Tags" in item:
                    mutelist["Accounts"][item["Accounts"]]["Checks"][item["Checks"]][
                        "Tags"
                    ] = item["Tags"]
                if "Exceptions" in item:
                    mutelist["Accounts"][item["Accounts"]]["Checks"][item["Checks"]][
                        "Exceptions"
                    ] = item["Exceptions"]
        else:
            with open(mutelist_path) as f:
                mutelist = yaml.safe_load(f)["Mutelist"]
        try:
            mutelist_schema.validate(mutelist)
        except Exception as error:
            logger.critical(
                f"{error.__class__.__name__} -- Mutelist YAML is malformed - {error}[{error.__traceback__.tb_lineno}]"
            )
            sys.exit(1)
        return mutelist
    except Exception as error:
        logger.critical(
            f"{error.__class__.__name__} -- {error}[{error.__traceback__.tb_lineno}]"
        )
        sys.exit(1)


def mutelist_findings(
    global_provider: Any,
    check_findings: list[Any],
):
    # Check if finding is muted
    for finding in check_findings:
        # TODO: Move this mapping to the execute_check function and pass that output to the mutelist and the report
        if global_provider.type == "aws":
            finding.muted = is_muted(
                global_provider.mutelist,
                global_provider.identity.account,
                finding.check_metadata.CheckID,
                finding.region,
                finding.resource_id,
                unroll_tags(finding.resource_tags),
            )
        elif global_provider.type == "azure":
            finding.muted = is_muted(
                global_provider.mutelist,
                finding.subscription,
                finding.check_metadata.CheckID,
                # TODO: add region to the findings when we add Azure Locations
                # finding.region,
                "",
                finding.resource_name,
                unroll_tags(finding.resource_tags),
            )
        elif global_provider.type == "gcp":
            finding.muted = is_muted(
                global_provider.mutelist,
                finding.project_id,
                finding.check_metadata.CheckID,
                finding.location,
                finding.resource_name,
                unroll_tags(finding.resource_tags),
            )
        elif global_provider.type == "kubernetes":
            finding.muted = is_muted(
                global_provider.mutelist,
                global_provider.identity.cluster,
                finding.check_metadata.CheckID,
                finding.namespace,
                finding.resource_name,
                unroll_tags(finding.resource_tags),
            )
    return check_findings


def is_muted(
    mutelist: dict,
    audited_account: str,
    check: str,
    finding_region: str,
    finding_resource: str,
    finding_tags,
):
    try:
        # By default is not muted
        is_finding_muted = False

        # We always check all the accounts present in the mutelist
        # if one mutes the finding we set the finding as muted
        for account in mutelist["Accounts"]:
            if account == audited_account or account == "*":
                if is_muted_in_check(
                    mutelist["Accounts"][account]["Checks"],
                    audited_account,
                    check,
                    finding_region,
                    finding_resource,
                    finding_tags,
                ):
                    is_finding_muted = True
                    break

        return is_finding_muted
    except Exception as error:
        logger.critical(
            f"{error.__class__.__name__} -- {error}[{error.__traceback__.tb_lineno}]"
        )
        sys.exit(1)


def is_muted_in_check(
    muted_checks,
    audited_account,
    check,
    finding_region,
    finding_resource,
    finding_tags,
):
    try:
        # Default value is not muted
        is_check_muted = False

        for muted_check, muted_check_info in muted_checks.items():
            # map lambda to awslambda
            muted_check = re.sub("^lambda", "awslambda", muted_check)

            check_match = (
                "*" == muted_check
                or check == muted_check
                or re.search(muted_check, check)
            )
            # Check if the finding is excepted
            exceptions = muted_check_info.get("Exceptions")
            if (
                is_excepted(
                    exceptions,
                    audited_account,
                    finding_region,
                    finding_resource,
                    finding_tags,
                )
                and check_match
            ):
                # Break loop and return default value since is excepted
                break

            muted_regions = muted_check_info.get("Regions")
            muted_resources = muted_check_info.get("Resources")
            muted_tags = muted_check_info.get("Tags", "*")
            # We need to set the muted_tags if None, "" or [], so the falsy helps
            if not muted_tags:
                muted_tags = "*"
            # If there is a *, it affects to all checks
            if check_match:
                muted_in_check = True
                muted_in_region = is_muted_in_region(muted_regions, finding_region)
                muted_in_resource = is_muted_in_resource(
                    muted_resources, finding_resource
                )
                muted_in_tags = is_muted_in_tags(muted_tags, finding_tags)

                # For a finding to be muted requires the following set to True:
                # - muted_in_check -> True
                # - muted_in_region -> True
                # - muted_in_tags -> True
                # - muted_in_resource -> True
                # - excepted -> False

                if (
                    muted_in_check
                    and muted_in_region
                    and muted_in_tags
                    and muted_in_resource
                ):
                    is_check_muted = True

        return is_check_muted
    except Exception as error:
        logger.critical(
            f"{error.__class__.__name__} -- {error}[{error.__traceback__.tb_lineno}]"
        )
        sys.exit(1)


def is_muted_in_region(
    mutelist_regions,
    finding_region,
):
    try:
        return __is_item_matched__(mutelist_regions, finding_region)
    except Exception as error:
        logger.critical(
            f"{error.__class__.__name__} -- {error}[{error.__traceback__.tb_lineno}]"
        )
        sys.exit(1)


def is_muted_in_tags(muted_tags, finding_tags):
    try:
        return __is_item_matched__(muted_tags, finding_tags)
    except Exception as error:
        logger.critical(
            f"{error.__class__.__name__} -- {error}[{error.__traceback__.tb_lineno}]"
        )
        sys.exit(1)


def is_muted_in_resource(muted_resources, finding_resource):
    try:
        return __is_item_matched__(muted_resources, finding_resource)

    except Exception as error:
        logger.critical(
            f"{error.__class__.__name__} -- {error}[{error.__traceback__.tb_lineno}]"
        )
        sys.exit(1)


def is_excepted(
    exceptions,
    audited_account,
    finding_region,
    finding_resource,
    finding_tags,
):
    """is_excepted returns True if the account, region, resource and tags are excepted"""
    try:
        excepted = False
        is_account_excepted = False
        is_region_excepted = False
        is_resource_excepted = False
        is_tag_excepted = False
        if exceptions:
            excepted_accounts = exceptions.get("Accounts", [])
            is_account_excepted = __is_item_matched__(
                excepted_accounts, audited_account
            )

            excepted_regions = exceptions.get("Regions", [])
            is_region_excepted = __is_item_matched__(excepted_regions, finding_region)

            excepted_resources = exceptions.get("Resources", [])
            is_resource_excepted = __is_item_matched__(
                excepted_resources, finding_resource
            )

            excepted_tags = exceptions.get("Tags", [])
            is_tag_excepted = __is_item_matched__(excepted_tags, finding_tags)

            if (
                not is_account_excepted
                and not is_region_excepted
                and not is_resource_excepted
                and not is_tag_excepted
            ):
                excepted = False
            elif (
                (is_account_excepted or not excepted_accounts)
                and (is_region_excepted or not excepted_regions)
                and (is_resource_excepted or not excepted_resources)
                and (is_tag_excepted or not excepted_tags)
            ):
                excepted = True
        return excepted
    except Exception as error:
        logger.critical(
            f"{error.__class__.__name__} -- {error}[{error.__traceback__.tb_lineno}]"
        )
        sys.exit(1)


def __is_item_matched__(matched_items, finding_items):
    """__is_item_matched__ return True if any of the matched_items are present in the finding_items, otherwise returns False."""
    try:
        is_item_matched = False
        if matched_items and (finding_items or finding_items == ""):
            for item in matched_items:
                if item == "*":
                    item = ".*"
                if re.search(item, finding_items):
                    is_item_matched = True
                    break
        return is_item_matched
    except Exception as error:
        logger.critical(
            f"{error.__class__.__name__} -- {error}[{error.__traceback__.tb_lineno}]"
        )
        sys.exit(1)



from py_ocsf_models.events.base_event import SeverityID, StatusID
from py_ocsf_models.events.findings.detection_finding import DetectionFinding
from py_ocsf_models.events.findings.detection_finding import (
    TypeID as DetectionFindingTypeID,
)
from py_ocsf_models.events.findings.finding import ActivityID, FindingInformation
from py_ocsf_models.objects.account import Account, TypeID
from py_ocsf_models.objects.cloud import Cloud
from py_ocsf_models.objects.container import Container
from py_ocsf_models.objects.group import Group
from py_ocsf_models.objects.metadata import Metadata
from py_ocsf_models.objects.organization import Organization
from py_ocsf_models.objects.product import Product
from py_ocsf_models.objects.remediation import Remediation
from py_ocsf_models.objects.resource_details import ResourceDetails

from prowler.lib.logger import logger
from prowler.lib.outputs.common_models import FindingOutput


def get_account_type_id_by_provider(provider: str) -> TypeID:
    type_id = TypeID.Other
    if provider == "aws":
        type_id = TypeID.AWS_Account
    elif provider == "azure":
        type_id = TypeID.Azure_AD_Account
    elif provider == "gcp":
        type_id = TypeID.GCP_Account
    return type_id


def get_finding_status_id(status: str, muted: bool) -> StatusID:
    status_id = StatusID.Other
    if status == "FAIL":
        status_id = StatusID.New
    if muted:
        status_id = StatusID.Suppressed
    return status_id


def fill_json_ocsf(finding_output: FindingOutput) -> DetectionFinding:
    try:
        finding_activity = ActivityID.Create
        cloud_account_type = get_account_type_id_by_provider(finding_output.provider)
        finding_severity = getattr(
            SeverityID, finding_output.severity.capitalize(), SeverityID.Unknown
        )
        finding_status = get_finding_status_id(
            finding_output.status, finding_output.muted
        )

        detection_finding = DetectionFinding(
            activity_id=finding_activity.value,
            activity_name=finding_activity.name,
            finding_info=FindingInformation(
                created_time=finding_output.timestamp,
                desc=finding_output.description,
                title=finding_output.check_title,
                uid=finding_output.finding_uid,
                product_uid="prowler",
            ),
            event_time=finding_output.timestamp,
            remediation=Remediation(
                desc=finding_output.remediation_recommendation_text,
                references=list(
                    filter(
                        None,
                        [
                            finding_output.remediation_code_nativeiac,
                            finding_output.remediation_code_terraform,
                            finding_output.remediation_code_cli,
                            finding_output.remediation_code_other,
                            finding_output.remediation_recommendation_url,
                        ],
                    )
                ),
            ),
            severity_id=finding_severity.value,
            severity=finding_severity.name,
            status_id=finding_status.value,
            status=finding_status.name,
            status_code=finding_output.status,
            status_detail=finding_output.status_extended,
            risk_details=finding_output.risk,
            resources=[
                ResourceDetails(
                    # TODO: Check labels for other providers
                    labels=(
                        finding_output.resource_tags.split(",")
                        if finding_output.resource_tags
                        else []
                    ),
                    name=finding_output.resource_name,
                    uid=finding_output.resource_uid,
                    group=Group(name=finding_output.service_name),
                    type=finding_output.resource_type,
                    # TODO: this should be included only if using the Cloud profile
                    cloud_partition=finding_output.partition,
                    region=finding_output.region,
                    data={"details": finding_output.resource_details},
                )
            ],
            metadata=Metadata(
                event_code=finding_output.check_id,
                product=Product(
                    name="Prowler",
                    vendor_name="Prowler",
                    version=finding_output.prowler_version,
                ),
            ),
            type_uid=DetectionFindingTypeID.Create,
            type_name=DetectionFindingTypeID.Create.name,
            unmapped={
                "check_type": finding_output.check_type,
                "related_url": finding_output.related_url,
                "categories": finding_output.categories,
                "depends_on": finding_output.depends_on,
                "related_to": finding_output.related_to,
                "notes": finding_output.notes,
                "compliance": finding_output.compliance,
            },
        )

        if finding_output.provider == "kubernetes":
            detection_finding.container = (
                Container(
                    name=finding_output.resource_name,
                    uid=finding_output.resource_uid,
                ),
            )
            # TODO: Get the PID of the namespace (we only have the name of the namespace)
            # detection_finding.namespace_pid=,
        else:
            detection_finding.cloud = Cloud(
                account=Account(
                    name=finding_output.account_name,
                    type_id=cloud_account_type.value,
                    type=cloud_account_type.name,
                    uid=finding_output.account_uid,
                    labels=finding_output.account_tags,
                ),
                org=Organization(
                    uid=finding_output.account_organization_uid,
                    name=finding_output.account_organization_name,
                ),
                provider=finding_output.provider,
                region=finding_output.region,
            )

        return detection_finding

    except Exception as error:
        logger.error(
            f"{error.__class__.__name__}[{error.__traceback__.tb_lineno}]: {error}"
        )



from datetime import datetime
from enum import Enum
from typing import Optional

from pydantic import BaseModel


class Status(str, Enum):
    PASS = "PASS"
    FAIL = "FAIL"
    MANUAL = "MANUAL"


class Severity(str, Enum):
    critical = "critical"
    high = "high"
    medium = "medium"
    low = "low"
    informational = "informational"


class CSVRow(BaseModel):
    """
    CSVRow generates a finding's output in CSV format.

    This is the base CSV output model for every provider.
    """

    auth_method: str
    timestamp: datetime
    account_uid: str
    # Optional since depends on permissions
    account_name: Optional[str]
    # Optional since depends on permissions
    account_email: Optional[str]
    # Optional since depends on permissions
    account_organization_uid: Optional[str]
    # Optional since depends on permissions
    account_organization: Optional[str]
    # Optional since depends on permissions
    account_tags: Optional[str]
    finding_uid: str
    provider: str
    check_id: str
    check_title: str
    check_type: str
    status: Status
    status_extended: str
    muted: bool = False
    service_name: str
    subservice_name: str
    severity: Severity
    resource_type: str
    resource_uid: str
    resource_name: str
    resource_details: str
    resource_tags: str
    # Only present for AWS and Azure
    partition: Optional[str]
    region: str
    description: str
    risk: str
    related_url: str
    remediation_recommendation_text: str
    remediation_recommendation_url: str
    remediation_code_nativeiac: str
    remediation_code_terraform: str
    remediation_code_cli: str
    remediation_code_other: str
    compliance: str
    categories: str
    depends_on: str
    related_to: str
    notes: str



from operator import attrgetter

from prowler.config.config import timestamp
from prowler.lib.logger import logger
from prowler.lib.outputs.common_models import FindingOutput
from prowler.lib.outputs.utils import unroll_list, unroll_tags
from prowler.lib.utils.utils import outputs_unix_timestamp


def get_provider_data_mapping(provider) -> dict:
    data = {}
    for generic_field, provider_field in provider.get_output_mapping.items():
        try:
            provider_value = attrgetter(provider_field)(provider)
            data[generic_field] = provider_value
        except AttributeError:
            data[generic_field] = ""
        except Exception as error:
            logger.error(
                f"{error.__class__.__name__}[{error.__traceback__.tb_lineno}]: {error}"
            )
    return data


def generate_provider_output(provider, finding, csv_data) -> FindingOutput:
    """
    generate_provider_output returns the provider's Finding output model
    """
    # TODO: we have to standardize this between the above mapping and the provider.get_output_mapping()
    try:
        if provider.type == "aws":
            # TODO: probably Organization UID is without the account id
            csv_data["auth_method"] = f"profile: {csv_data['auth_method']}"
            csv_data["resource_name"] = finding.resource_id
            csv_data["resource_uid"] = finding.resource_arn
            csv_data["region"] = finding.region

        elif provider.type == "azure":
            # TODO: we should show the authentication method used I think
            csv_data["auth_method"] = (
                f"{provider.identity.identity_type}: {provider.identity.identity_id}"
            )
            # Get the first tenant domain ID, just in case
            csv_data["account_organization_uid"] = csv_data["account_organization_uid"][
                0
            ]
            csv_data["account_uid"] = (
                csv_data["account_organization_uid"]
                if "Tenant:" in finding.subscription
                else provider.identity.subscriptions[finding.subscription]
            )
            csv_data["account_name"] = finding.subscription
            csv_data["resource_name"] = finding.resource_name
            csv_data["resource_uid"] = finding.resource_id
            csv_data["region"] = finding.location

        elif provider.type == "gcp":
            csv_data["auth_method"] = f"Principal: {csv_data['auth_method']}"
            csv_data["account_uid"] = provider.projects[finding.project_id].id
            csv_data["account_name"] = provider.projects[finding.project_id].name
            csv_data["account_tags"] = provider.projects[finding.project_id].labels
            csv_data["resource_name"] = finding.resource_name
            csv_data["resource_uid"] = finding.resource_id
            csv_data["region"] = finding.location

            if (
                provider.projects
                and finding.project_id in provider.projects
                and getattr(provider.projects[finding.project_id], "organization")
            ):
                csv_data["account_organization_uid"] = provider.projects[
                    finding.project_id
                ].organization.id
                # TODO: for now is None since we don't retrieve that data
                csv_data["account_organization"] = provider.projects[
                    finding.project_id
                ].organization.display_name

        elif provider.type == "kubernetes":
            if provider.identity.context == "In-Cluster":
                csv_data["auth_method"] = "in-cluster"
            else:
                csv_data["auth_method"] = "kubeconfig"
            csv_data["resource_name"] = finding.resource_name
            csv_data["resource_uid"] = finding.resource_id
            csv_data["account_name"] = f"context: {provider.identity.context}"
            csv_data["region"] = f"namespace: {finding.namespace}"

        # Finding Unique ID
        # TODO: move this to a function
        # TODO: in Azure, GCP and K8s there are fidings without resource_name
        csv_data["finding_uid"] = (
            f"prowler-{provider.type}-{finding.check_metadata.CheckID}-{csv_data['account_uid']}-{csv_data['region']}-{csv_data['resource_name']}"
        )

        finding_output = FindingOutput(**csv_data)

    except Exception as error:
        logger.error(
            f"{error.__class__.__name__}[{error.__traceback__.tb_lineno}]: {error}"
        )
    else:
        return finding_output


# TODO: add test for outputs_unix_timestamp
def fill_common_finding_data(finding: dict, unix_timestamp: bool) -> dict:
    finding_data = {
        "timestamp": outputs_unix_timestamp(unix_timestamp, timestamp),
        "check_id": finding.check_metadata.CheckID,
        "check_title": finding.check_metadata.CheckTitle,
        "check_type": ",".join(finding.check_metadata.CheckType),
        "status": finding.status,
        "status_extended": finding.status_extended,
        "muted": finding.muted,
        "service_name": finding.check_metadata.ServiceName,
        "subservice_name": finding.check_metadata.SubServiceName,
        "severity": finding.check_metadata.Severity,
        "resource_type": finding.check_metadata.ResourceType,
        "resource_details": finding.resource_details,
        "resource_tags": unroll_tags(finding.resource_tags),
        "description": finding.check_metadata.Description,
        "risk": finding.check_metadata.Risk,
        "related_url": finding.check_metadata.RelatedUrl,
        "remediation_recommendation_text": (
            finding.check_metadata.Remediation.Recommendation.Text
        ),
        "remediation_recommendation_url": (
            finding.check_metadata.Remediation.Recommendation.Url
        ),
        "remediation_code_nativeiac": (
            finding.check_metadata.Remediation.Code.NativeIaC
        ),
        "remediation_code_terraform": (
            finding.check_metadata.Remediation.Code.Terraform
        ),
        "remediation_code_cli": (finding.check_metadata.Remediation.Code.CLI),
        "remediation_code_other": (finding.check_metadata.Remediation.Code.Other),
        "categories": unroll_list(finding.check_metadata.Categories),
        "depends_on": unroll_list(finding.check_metadata.DependsOn),
        "related_to": unroll_list(finding.check_metadata.RelatedTo),
        "notes": finding.check_metadata.Notes,
    }
    return finding_data



from prowler.config.config import timestamp
from prowler.lib.outputs.compliance.models import Check_Output_CSV_GCP_CIS
from prowler.lib.outputs.csv.csv import generate_csv_fields
from prowler.lib.utils.utils import outputs_unix_timestamp


def generate_compliance_row_cis_gcp(
    finding, compliance, requirement, attribute, output_options
):
    compliance_row = Check_Output_CSV_GCP_CIS(
        Provider=finding.check_metadata.Provider,
        Description=compliance.Description,
        ProjectId=finding.project_id,
        Location=finding.location.lower(),
        AssessmentDate=outputs_unix_timestamp(output_options.unix_timestamp, timestamp),
        Requirements_Id=requirement.Id,
        Requirements_Description=requirement.Description,
        Requirements_Attributes_Section=attribute.Section,
        Requirements_Attributes_Profile=attribute.Profile,
        Requirements_Attributes_AssessmentStatus=attribute.AssessmentStatus,
        Requirements_Attributes_Description=attribute.Description,
        Requirements_Attributes_RationaleStatement=attribute.RationaleStatement,
        Requirements_Attributes_ImpactStatement=attribute.ImpactStatement,
        Requirements_Attributes_RemediationProcedure=attribute.RemediationProcedure,
        Requirements_Attributes_AuditProcedure=attribute.AuditProcedure,
        Requirements_Attributes_AdditionalInformation=attribute.AdditionalInformation,
        Requirements_Attributes_References=attribute.References,
        Status=finding.status,
        StatusExtended=finding.status_extended,
        ResourceId=finding.resource_id,
        ResourceName=finding.resource_name,
        CheckId=finding.check_metadata.CheckID,
        Muted=finding.muted,
    )
    csv_header = generate_csv_fields(Check_Output_CSV_GCP_CIS)

    return compliance_row, csv_header



def unroll_list(listed_items: list, separator: str = "|"):
    unrolled_items = ""
    if listed_items:
        for item in listed_items:
            if not unrolled_items:
                unrolled_items = f"{item}"
            else:
                if separator == "|":
                    unrolled_items = f"{unrolled_items} {separator} {item}"
                else:
                    unrolled_items = f"{unrolled_items}{separator} {item}"

    return unrolled_items


def unroll_tags(tags: list):
    unrolled_items = ""
    separator = "|"
    if tags and tags != [{}] and tags != [None]:
        for item in tags:
            # Check if there are tags in list
            if isinstance(item, dict):
                for key, value in item.items():
                    if not unrolled_items:
                        # Check the pattern of tags (Key:Value or Key:key/Value:value)
                        if "Key" != key and "Value" != key:
                            unrolled_items = f"{key}={value}"
                        else:
                            if "Key" == key:
                                unrolled_items = f"{value}="
                            else:
                                unrolled_items = f"{value}"
                    else:
                        if "Key" != key and "Value" != key:
                            unrolled_items = (
                                f"{unrolled_items} {separator} {key}={value}"
                            )
                        else:
                            if "Key" == key:
                                unrolled_items = (
                                    f"{unrolled_items} {separator} {value}="
                                )
                            else:
                                unrolled_items = f"{unrolled_items}{value}"
            elif not unrolled_items:
                unrolled_items = f"{item}"
            else:
                unrolled_items = f"{unrolled_items} {separator} {item}"

    return unrolled_items


def unroll_dict(dict: dict):
    unrolled_items = ""
    separator = "|"
    for key, value in dict.items():
        if isinstance(value, list):
            value = ", ".join(value)
        if not unrolled_items:
            unrolled_items = f"{key}: {value}"
        else:
            unrolled_items = f"{unrolled_items} {separator} {key}: {value}"

    return unrolled_items


def unroll_dict_to_list(dict: dict):
    dict_list = []
    for key, value in dict.items():
        if isinstance(value, list):
            value = ", ".join(value)
            dict_list.append(f"{key}: {value}")
        else:
            dict_list.append(f"{key}: {value}")

    return dict_list


def parse_json_tags(tags: list):
    dict_tags = {}
    if tags and tags != [{}] and tags != [None]:
        for tag in tags:
            if "Key" in tag and "Value" in tag:
                dict_tags[tag["Key"]] = tag["Value"]
            else:
                dict_tags.update(tag)

    return dict_tags


def parse_html_string(str: str):
    string = ""
    for elem in str.split(" | "):
        if elem:
            string += f"\n&#x2022;{elem}\n"

    return string



from prowler.config.config import timestamp_utc
from prowler.lib.logger import logger
from prowler.lib.outputs.compliance.compliance import get_check_compliance
from prowler.lib.outputs.json_asff.models import (
    Check_Output_JSON_ASFF,
    Compliance,
    ProductFields,
    Recommendation,
    Remediation,
    Resource,
    Severity,
)
from prowler.lib.utils.utils import hash_sha512


def generate_json_asff_status(status: str, muted: bool = False) -> str:
    json_asff_status = ""
    if muted:
        # Per AWS Security Hub "MUTED" is not a valid status
        # https://docs.aws.amazon.com/securityhub/1.0/APIReference/API_Compliance.html
        json_asff_status = "WARNING"
    else:
        if status == "PASS":
            json_asff_status = "PASSED"
        elif status == "FAIL":
            json_asff_status = "FAILED"
        else:
            json_asff_status = "NOT_AVAILABLE"

    return json_asff_status


def generate_json_asff_resource_tags(tags):
    try:
        resource_tags = {}
        if tags and tags != [None]:
            for tag in tags:
                if "Key" in tag and "Value" in tag:
                    resource_tags[tag["Key"]] = tag["Value"]
                else:
                    resource_tags.update(tag)
            if len(resource_tags) == 0:
                return None
        else:
            return None
        return resource_tags
    except Exception as error:
        logger.error(
            f"{error.__class__.__name__}[{error.__traceback__.tb_lineno}]: {error}"
        )


def fill_json_asff(provider, finding):
    """
    Fill the finding's output in JSON ASFF format.

    Parameters:
    - provider: The provider object containing information about the provider (e.g., AWS) and the output options object containing information about the desired output format.
    - finding: The finding object containing information about the specific finding.

    Returns:
    - finding_output: The filled finding's output in JSON ASFF format.
    """

    try:
        # Check if there are no resources in the finding
        if finding.resource_arn == "":
            if finding.resource_id == "":
                finding.resource_id = "NONE_PROVIDED"
            finding.resource_arn = finding.resource_id

        timestamp = timestamp_utc.strftime("%Y-%m-%dT%H:%M:%SZ")
        resource_tags = generate_json_asff_resource_tags(finding.resource_tags)

        # Iterate for each compliance framework
        compliance_summary = []
        associated_standards = []
        check_compliance = get_check_compliance(
            finding, provider.type, provider.output_options
        )
        for key, value in check_compliance.items():
            if (
                len(associated_standards) < 20
            ):  # AssociatedStandards should NOT have more than 20 items
                associated_standards.append({"StandardsId": key})
                item = f"{key} {' '.join(value)}"
                if len(item) > 64:
                    item = item[0:63]
                compliance_summary.append(item)

        # Ensures finding_status matches allowed values in ASFF
        finding_status = generate_json_asff_status(finding.status, finding.muted)

        json_asff_output = Check_Output_JSON_ASFF(
            # The following line cannot be changed because it is the format we use to generate unique findings for AWS Security Hub
            # If changed some findings could be lost because the unique identifier will be different
            # TODO: get this from the provider output
            Id=f"prowler-{finding.check_metadata.CheckID}-{provider.identity.account}-{finding.region}-{hash_sha512(finding.resource_id)}",
            ProductArn=f"arn:{provider.identity.partition}:securityhub:{finding.region}::product/prowler/prowler",
            ProductFields=ProductFields(
                ProwlerResourceName=finding.resource_arn,
            ),
            GeneratorId="prowler-" + finding.check_metadata.CheckID,
            AwsAccountId=provider.identity.account,
            Types=finding.check_metadata.CheckType,
            FirstObservedAt=timestamp,
            UpdatedAt=timestamp,
            CreatedAt=timestamp,
            Severity=Severity(Label=finding.check_metadata.Severity.upper()),
            Title=finding.check_metadata.CheckTitle,
            Description=finding.status_extended,
            Resources=[
                Resource(
                    Id=finding.resource_arn,
                    Type=finding.check_metadata.ResourceType,
                    Partition=provider.identity.partition,
                    Region=finding.region,
                    Tags=resource_tags,
                )
            ],
            Compliance=Compliance(
                Status=finding_status,
                AssociatedStandards=associated_standards,
                RelatedRequirements=compliance_summary,
            ),
            Remediation=Remediation(
                Recommendation=Recommendation(
                    Text=finding.check_metadata.Remediation.Recommendation.Text,
                    Url=finding.check_metadata.Remediation.Recommendation.Url,
                )
            ),
        )
        return json_asff_output
    except Exception as error:
        logger.error(
            f"{error.__class__.__name__}[{error.__traceback__.tb_lineno}]: {error}"
        )



import sys

from prowler.lib.check.models import Check_Report
from prowler.lib.logger import logger
from prowler.lib.outputs.compliance.aws_well_architected_framework import (
    write_compliance_row_aws_well_architected_framework,
)
from prowler.lib.outputs.compliance.cis import get_cis_table, write_compliance_row_cis
from prowler.lib.outputs.compliance.ens_rd2022_aws import (
    get_ens_rd2022_aws_table,
    write_compliance_row_ens_rd2022_aws,
)
from prowler.lib.outputs.compliance.generic import (
    get_generic_compliance_table,
    write_compliance_row_generic,
)
from prowler.lib.outputs.compliance.iso27001_2013_aws import (
    write_compliance_row_iso27001_2013_aws,
)
from prowler.lib.outputs.compliance.mitre_attack.mitre_attack import (
    get_mitre_attack_table,
    write_compliance_row_mitre_attack,
)


def add_manual_controls(
    output_options, provider, file_descriptors, input_compliance_frameworks
):
    try:
        # Check if MANUAL control was already added to output
        if "manual_check" in output_options.bulk_checks_metadata:
            manual_finding = Check_Report(
                output_options.bulk_checks_metadata["manual_check"].json()
            )
            manual_finding.status = "MANUAL"
            manual_finding.status_extended = "Manual check"
            manual_finding.resource_id = "manual_check"
            manual_finding.resource_name = "Manual check"
            manual_finding.region = ""
            manual_finding.location = ""
            manual_finding.project_id = ""
            manual_finding.subscription = ""
            manual_finding.namespace = ""
            fill_compliance(
                output_options,
                manual_finding,
                provider,
                file_descriptors,
                input_compliance_frameworks,
            )
            del output_options.bulk_checks_metadata["manual_check"]
    except Exception as error:
        logger.error(
            f"{error.__class__.__name__}[{error.__traceback__.tb_lineno}]: {error}"
        )


def get_check_compliance_frameworks_in_input(
    check_id, bulk_checks_metadata, input_compliance_frameworks
):
    """get_check_compliance_frameworks_in_input returns a list of Compliance for the given check if the compliance framework is present in the input compliance to execute"""
    check_compliances = []
    if bulk_checks_metadata and bulk_checks_metadata[check_id]:
        for compliance in bulk_checks_metadata[check_id].Compliance:
            compliance_name = ""
            if compliance.Version:
                compliance_name = (
                    compliance.Framework.lower()
                    + "_"
                    + compliance.Version.lower()
                    + "_"
                    + compliance.Provider.lower()
                )
            else:
                compliance_name = (
                    compliance.Framework.lower() + "_" + compliance.Provider.lower()
                )
            if compliance_name.replace("-", "_") in input_compliance_frameworks:
                check_compliances.append(compliance)
    return check_compliances


def fill_compliance(
    output_options, finding, provider, file_descriptors, input_compliance_frameworks
):
    try:
        # We have to retrieve all the check's compliance requirements and get the ones matching with the input ones
        check_compliances = get_check_compliance_frameworks_in_input(
            finding.check_metadata.CheckID,
            output_options.bulk_checks_metadata,
            input_compliance_frameworks,
        )

        for compliance in check_compliances:
            if compliance.Framework == "ENS" and compliance.Version == "RD2022":
                write_compliance_row_ens_rd2022_aws(
                    file_descriptors, finding, compliance, output_options, provider
                )

            elif compliance.Framework == "CIS":
                write_compliance_row_cis(
                    file_descriptors,
                    finding,
                    compliance,
                    output_options,
                    provider,
                    input_compliance_frameworks,
                )

            elif (
                "AWS-Well-Architected-Framework" in compliance.Framework
                and compliance.Provider == "AWS"
            ):
                write_compliance_row_aws_well_architected_framework(
                    file_descriptors, finding, compliance, output_options, provider
                )

            elif (
                compliance.Framework == "ISO27001"
                and compliance.Version == "2013"
                and compliance.Provider == "AWS"
            ):
                write_compliance_row_iso27001_2013_aws(
                    file_descriptors, finding, compliance, output_options, provider
                )

            elif compliance.Framework == "MITRE-ATTACK" and compliance.Version == "":
                write_compliance_row_mitre_attack(
                    file_descriptors, finding, compliance, provider
                )

            else:
                write_compliance_row_generic(
                    file_descriptors, finding, compliance, output_options, provider
                )

    except Exception as error:
        logger.error(
            f"{error.__class__.__name__}[{error.__traceback__.tb_lineno}]: {error}"
        )


def display_compliance_table(
    findings: list,
    bulk_checks_metadata: dict,
    compliance_framework: str,
    output_filename: str,
    output_directory: str,
    compliance_overview: bool,
):
    try:
        if "ens_rd2022_aws" == compliance_framework:
            get_ens_rd2022_aws_table(
                findings,
                bulk_checks_metadata,
                compliance_framework,
                output_filename,
                output_directory,
                compliance_overview,
            )
        elif "cis_" in compliance_framework:
            get_cis_table(
                findings,
                bulk_checks_metadata,
                compliance_framework,
                output_filename,
                output_directory,
                compliance_overview,
            )
        elif "mitre_attack" in compliance_framework:
            get_mitre_attack_table(
                findings,
                bulk_checks_metadata,
                compliance_framework,
                output_filename,
                output_directory,
                compliance_overview,
            )
        else:
            get_generic_compliance_table(
                findings,
                bulk_checks_metadata,
                compliance_framework,
                output_filename,
                output_directory,
                compliance_overview,
            )
    except Exception as error:
        logger.critical(
            f"{error.__class__.__name__}:{error.__traceback__.tb_lineno} -- {error}"
        )
        sys.exit(1)


def get_check_compliance(finding, provider_type, output_options) -> dict:
    """get_check_compliance returns a map with the compliance framework as key and the requirements where the finding's check is present.

        Example:

    {
        "CIS-1.4": ["2.1.3"],
        "CIS-1.5": ["2.1.3"],
    }
    """
    try:
        check_compliance = {}
        # We have to retrieve all the check's compliance requirements
        if finding.check_metadata.CheckID in output_options.bulk_checks_metadata:
            for compliance in output_options.bulk_checks_metadata[
                finding.check_metadata.CheckID
            ].Compliance:
                compliance_fw = compliance.Framework
                if compliance.Version:
                    compliance_fw = f"{compliance_fw}-{compliance.Version}"
                # compliance.Provider == "Azure" or "Kubernetes"
                # provider_type == "azure" or "kubernetes"
                if compliance.Provider.upper() == provider_type.upper():
                    if compliance_fw not in check_compliance:
                        check_compliance[compliance_fw] = []
                    for requirement in compliance.Requirements:
                        check_compliance[compliance_fw].append(requirement.Id)
        return check_compliance
    except Exception as error:
        logger.critical(
            f"{error.__class__.__name__}[{error.__traceback__.tb_lineno}] -- {error}"
        )
        sys.exit(1)



import json
from csv import DictWriter

from colorama import Fore, Style

from prowler.config.config import available_compliance_frameworks, orange_color
from prowler.lib.logger import logger
from prowler.lib.outputs.common import (
    fill_common_finding_data,
    generate_provider_output,
    get_provider_data_mapping,
)
from prowler.lib.outputs.common_models import FindingOutput
from prowler.lib.outputs.compliance.compliance import (
    add_manual_controls,
    fill_compliance,
    get_check_compliance,
)
from prowler.lib.outputs.csv.csv import generate_csv_fields
from prowler.lib.outputs.file_descriptors import fill_file_descriptors
from prowler.lib.outputs.html.html import fill_html
from prowler.lib.outputs.json_asff.json_asff import fill_json_asff
from prowler.lib.outputs.json_ocsf.json_ocsf import fill_json_ocsf
from prowler.lib.outputs.utils import unroll_dict, unroll_list


def stdout_report(finding, color, verbose, status, fix):
    if finding.check_metadata.Provider == "aws":
        details = finding.region
    if finding.check_metadata.Provider == "azure":
        details = finding.check_metadata.ServiceName
    if finding.check_metadata.Provider == "gcp":
        details = finding.location.lower()
    if finding.check_metadata.Provider == "kubernetes":
        details = finding.namespace.lower()

    if (verbose or fix) and (not status or finding.status in status):
        if finding.muted:
            print(
                f"\t{color}MUTED ({finding.status}){Style.RESET_ALL} {details}: {finding.status_extended}"
            )
        else:
            print(
                f"\t{color}{finding.status}{Style.RESET_ALL} {details}: {finding.status_extended}"
            )


def report(check_findings, provider):
    try:
        output_options = provider.output_options
        file_descriptors = {}
        if check_findings:
            # TO-DO Generic Function
            if provider.type == "aws":
                check_findings.sort(key=lambda x: x.region)

            if provider.type == "azure":
                check_findings.sort(key=lambda x: x.subscription)

            # Generate the required output files
            if output_options.output_modes and not output_options.fixer:
                # We have to create the required output files
                file_descriptors = fill_file_descriptors(
                    output_options.output_modes,
                    output_options.output_directory,
                    output_options.output_filename,
                    provider,
                )

            for finding in check_findings:
                # Print findings by stdout
                color = set_report_color(finding.status, finding.muted)
                stdout_report(
                    finding,
                    color,
                    output_options.verbose,
                    output_options.status,
                    output_options.fixer,
                )

                if file_descriptors:
                    # Check if --status is enabled and if the filter applies
                    if (
                        not output_options.status
                        or finding.status in output_options.status
                    ):
                        input_compliance_frameworks = list(
                            set(output_options.output_modes).intersection(
                                available_compliance_frameworks
                            )
                        )
                        fill_compliance(
                            output_options,
                            finding,
                            provider,
                            file_descriptors,
                            input_compliance_frameworks,
                        )

                        add_manual_controls(
                            output_options,
                            provider,
                            file_descriptors,
                            input_compliance_frameworks,
                        )

                        # AWS specific outputs
                        if finding.check_metadata.Provider == "aws":
                            if "json-asff" in file_descriptors:
                                # Initialize this field using the class within fill_json_asff not here
                                json_asff_finding = fill_json_asff(provider, finding)

                                json.dump(
                                    json_asff_finding.dict(exclude_none=True),
                                    file_descriptors["json-asff"],
                                    indent=4,
                                )
                                file_descriptors["json-asff"].write(",")

                        # Common Output Data
                        provider_data_mapping = get_provider_data_mapping(provider)
                        common_finding_data = fill_common_finding_data(
                            finding, output_options.unix_timestamp
                        )
                        csv_data = {}
                        csv_data.update(provider_data_mapping)
                        csv_data.update(common_finding_data)
                        csv_data["compliance"] = get_check_compliance(
                            finding, provider.type, output_options
                        )
                        finding_output = generate_provider_output(
                            provider, finding, csv_data
                        )
                        # JSON
                        if "json-ocsf" in file_descriptors:
                            detection_finding = fill_json_ocsf(finding_output)
                            # print(file)
                            file_descriptors["json-ocsf"].write(
                                detection_finding.json(exclude_none=True, indent=4)
                            )
                            file_descriptors["json-ocsf"].write(",")

                        if "html" in file_descriptors:
                            fill_html(file_descriptors["html"], finding_output)

                        # CSV
                        if "csv" in file_descriptors:
                            finding_output.compliance = unroll_dict(
                                finding_output.compliance
                            )
                            finding_output.account_tags = unroll_list(
                                finding_output.account_tags, ","
                            )
                            csv_writer = DictWriter(
                                file_descriptors["csv"],
                                fieldnames=generate_csv_fields(FindingOutput),
                                delimiter=";",
                            )

                            csv_writer.writerow(finding_output.dict())

        else:  # No service resources in the whole account
            color = set_report_color("MANUAL")
            if output_options.verbose:
                print(f"\t{color}INFO{Style.RESET_ALL} There are no resources")
        # Separator between findings and bar
        if output_options.verbose:
            print()
        if file_descriptors:
            # Close all file descriptors
            for file_descriptor in file_descriptors:
                file_descriptors.get(file_descriptor).close()
    except Exception as error:
        logger.error(
            f"{error.__class__.__name__}[{error.__traceback__.tb_lineno}]: {error}"
        )


def set_report_color(status: str, muted: bool = False) -> str:
    """Return the color for a give result status"""
    color = ""
    if muted:
        color = orange_color
    elif status == "PASS":
        color = Fore.GREEN
    elif status == "FAIL":
        color = Fore.RED
    elif status == "MANUAL":
        color = Fore.YELLOW
    else:
        raise Exception("Invalid Report Status. Must be PASS, FAIL or MANUAL.")
    return color


def extract_findings_statistics(findings: list) -> dict:
    """
    extract_findings_statistics takes a list of findings and returns the following dict with the aggregated statistics
    {
        "total_pass": 0,
        "total_fail": 0,
        "resources_count": 0,
        "findings_count": 0,
    }
    """
    logger.info("Extracting audit statistics...")
    stats = {}
    total_pass = 0
    total_fail = 0
    resources = set()
    findings_count = 0
    all_fails_are_muted = True

    for finding in findings:
        # Save the resource_id
        resources.add(finding.resource_id)
        if finding.status == "PASS":
            total_pass += 1
            findings_count += 1
        if finding.status == "FAIL":
            total_fail += 1
            findings_count += 1
            if not finding.muted and all_fails_are_muted:
                all_fails_are_muted = False

    stats["total_pass"] = total_pass
    stats["total_fail"] = total_fail
    stats["resources_count"] = len(resources)
    stats["findings_count"] = findings_count
    stats["all_fails_are_muted"] = all_fails_are_muted

    return stats



from typing import Optional

from pydantic import BaseModel


# TODO: move this to outputs/<compliance>/models.py
class Check_Output_CSV_ENS_RD2022(BaseModel):
    """
    Check_Output_CSV_ENS_RD2022 generates a finding's output in CSV ENS RD2022 format.
    """

    Provider: str
    Description: str
    AccountId: str
    Region: str
    AssessmentDate: str
    Requirements_Id: str
    Requirements_Description: str
    Requirements_Attributes_IdGrupoControl: str
    Requirements_Attributes_Marco: str
    Requirements_Attributes_Categoria: str
    Requirements_Attributes_DescripcionControl: str
    Requirements_Attributes_Nivel: str
    Requirements_Attributes_Tipo: str
    Requirements_Attributes_Dimensiones: str
    Requirements_Attributes_ModoEjecucion: str
    Requirements_Attributes_Dependencias: str
    Status: str
    StatusExtended: str
    ResourceId: str
    CheckId: str
    Muted: bool


class Check_Output_CSV_AWS_CIS(BaseModel):
    """
    Check_Output_CSV_CIS generates a finding's output in CSV CIS format.
    """

    Provider: str
    Description: str
    AccountId: str
    Region: str
    AssessmentDate: str
    Requirements_Id: str
    Requirements_Description: str
    Requirements_Attributes_Section: str
    Requirements_Attributes_Profile: str
    Requirements_Attributes_AssessmentStatus: str
    Requirements_Attributes_Description: str
    Requirements_Attributes_RationaleStatement: str
    Requirements_Attributes_ImpactStatement: str
    Requirements_Attributes_RemediationProcedure: str
    Requirements_Attributes_AuditProcedure: str
    Requirements_Attributes_AdditionalInformation: str
    Requirements_Attributes_References: str
    Status: str
    StatusExtended: str
    ResourceId: str
    CheckId: str
    Muted: bool


class Check_Output_CSV_AZURE_CIS(BaseModel):
    """
    Check_Output_CSV_CIS generates a finding's output in CSV CIS format.
    """

    Provider: str
    Description: str
    Subscription: str
    AssessmentDate: str
    Requirements_Id: str
    Requirements_Description: str
    Requirements_Attributes_Section: str
    Requirements_Attributes_Profile: str
    Requirements_Attributes_AssessmentStatus: str
    Requirements_Attributes_Description: str
    Requirements_Attributes_RationaleStatement: str
    Requirements_Attributes_ImpactStatement: str
    Requirements_Attributes_RemediationProcedure: str
    Requirements_Attributes_AuditProcedure: str
    Requirements_Attributes_AdditionalInformation: str
    Requirements_Attributes_DefaultValue: str
    Requirements_Attributes_References: str
    Status: str
    StatusExtended: str
    ResourceId: str
    ResourceName: str
    CheckId: str
    Muted: bool


class Check_Output_CSV_GCP_CIS(BaseModel):
    """
    Check_Output_CSV_CIS generates a finding's output in CSV CIS format.
    """

    Provider: str
    Description: str
    ProjectId: str
    Location: str
    AssessmentDate: str
    Requirements_Id: str
    Requirements_Description: str
    Requirements_Attributes_Section: str
    Requirements_Attributes_Profile: str
    Requirements_Attributes_AssessmentStatus: str
    Requirements_Attributes_Description: str
    Requirements_Attributes_RationaleStatement: str
    Requirements_Attributes_ImpactStatement: str
    Requirements_Attributes_RemediationProcedure: str
    Requirements_Attributes_AuditProcedure: str
    Requirements_Attributes_AdditionalInformation: str
    Requirements_Attributes_References: str
    Status: str
    StatusExtended: str
    ResourceId: str
    ResourceName: str
    CheckId: str
    Muted: bool


class Check_Output_CSV_KUBERNETES_CIS(BaseModel):
    """
    Check_Output_CSV_CIS generates a finding's output in CSV CIS format.
    """

    Provider: str
    Description: str
    Context: str
    Namespace: str
    AssessmentDate: str
    Requirements_Id: str
    Requirements_Description: str
    Requirements_Attributes_Section: str
    Requirements_Attributes_Profile: str
    Requirements_Attributes_AssessmentStatus: str
    Requirements_Attributes_Description: str
    Requirements_Attributes_RationaleStatement: str
    Requirements_Attributes_ImpactStatement: str
    Requirements_Attributes_RemediationProcedure: str
    Requirements_Attributes_AuditProcedure: str
    Requirements_Attributes_AdditionalInformation: str
    Requirements_Attributes_References: str
    Requirements_Attributes_DefaultValue: str
    Status: str
    StatusExtended: str
    ResourceId: str
    CheckId: str
    Muted: bool


class Check_Output_CSV_Generic_Compliance(BaseModel):
    """
    Check_Output_CSV_Generic_Compliance generates a finding's output in CSV Generic Compliance format.
    """

    Provider: str
    Description: str
    AccountId: str
    Region: str
    AssessmentDate: str
    Requirements_Id: str
    Requirements_Description: str
    Requirements_Attributes_Section: Optional[str]
    Requirements_Attributes_SubSection: Optional[str]
    Requirements_Attributes_SubGroup: Optional[str]
    Requirements_Attributes_Service: Optional[str]
    Requirements_Attributes_Type: Optional[str]
    Status: str
    StatusExtended: str
    ResourceId: str
    CheckId: str
    Muted: bool


class Check_Output_CSV_AWS_Well_Architected(BaseModel):
    """
    Check_Output_CSV_AWS_Well_Architected generates a finding's output in CSV AWS Well Architected Compliance format.
    """

    Provider: str
    Description: str
    AccountId: str
    Region: str
    AssessmentDate: str
    Requirements_Attributes_Name: str
    Requirements_Attributes_WellArchitectedQuestionId: str
    Requirements_Attributes_WellArchitectedPracticeId: str
    Requirements_Attributes_Section: str
    Requirements_Attributes_SubSection: Optional[str]
    Requirements_Attributes_LevelOfRisk: str
    Requirements_Attributes_AssessmentMethod: str
    Requirements_Attributes_Description: str
    Requirements_Attributes_ImplementationGuidanceUrl: str
    Status: str
    StatusExtended: str
    ResourceId: str
    CheckId: str
    Muted: bool


class Check_Output_CSV_AWS_ISO27001_2013(BaseModel):
    """
    Check_Output_CSV_AWS_ISO27001_2013 generates a finding's output in CSV AWS ISO27001 Compliance format.
    """

    Provider: str
    Description: str
    AccountId: str
    Region: str
    AssessmentDate: str
    Requirements_Attributes_Category: str
    Requirements_Attributes_Objetive_ID: str
    Requirements_Attributes_Objetive_Name: str
    Requirements_Attributes_Check_Summary: str
    Status: str
    StatusExtended: str
    ResourceId: str
    CheckId: str
    Muted: bool



from csv import DictWriter
from typing import Any


def write_csv(file_descriptor, headers, row):
    csv_writer = DictWriter(
        file_descriptor,
        fieldnames=headers,
        delimiter=";",
    )
    csv_writer.writerow(row.__dict__)


def generate_csv_fields(format: Any) -> list[str]:
    """Generates the CSV headers for the given class"""
    csv_fields = []
    # __fields__ is always available in the Pydantic's BaseModel class
    for field in format.__dict__.get("__fields__").keys():
        csv_fields.append(field)
    return csv_fields



from csv import DictWriter

from prowler.config.config import timestamp
from prowler.lib.outputs.compliance.models import Check_Output_CSV_AWS_ISO27001_2013
from prowler.lib.outputs.csv.csv import generate_csv_fields
from prowler.lib.utils.utils import outputs_unix_timestamp


def write_compliance_row_iso27001_2013_aws(
    file_descriptors, finding, compliance, output_options, provider
):
    compliance_output = compliance.Framework
    if compliance.Version != "":
        compliance_output += "_" + compliance.Version
    if compliance.Provider != "":
        compliance_output += "_" + compliance.Provider

    compliance_output = compliance_output.lower().replace("-", "_")
    csv_header = generate_csv_fields(Check_Output_CSV_AWS_ISO27001_2013)
    csv_writer = DictWriter(
        file_descriptors[compliance_output],
        fieldnames=csv_header,
        delimiter=";",
    )
    for requirement in compliance.Requirements:
        requirement_description = requirement.Description
        requirement_id = requirement.Id
        requirement_name = requirement.Name
        for attribute in requirement.Attributes:
            compliance_row = Check_Output_CSV_AWS_ISO27001_2013(
                Provider=finding.check_metadata.Provider,
                Description=compliance.Description,
                AccountId=provider.identity.account,
                Region=finding.region,
                AssessmentDate=outputs_unix_timestamp(
                    output_options.unix_timestamp, timestamp
                ),
                Requirements_Id=requirement_id,
                Requirements_Name=requirement_name,
                Requirements_Description=requirement_description,
                Requirements_Attributes_Category=attribute.Category,
                Requirements_Attributes_Objetive_ID=attribute.Objetive_ID,
                Requirements_Attributes_Objetive_Name=attribute.Objetive_Name,
                Requirements_Attributes_Check_Summary=attribute.Check_Summary,
                Status=finding.status,
                StatusExtended=finding.status_extended,
                ResourceId=finding.resource_id,
                CheckId=finding.check_metadata.CheckID,
                Muted=finding.muted,
            )

            csv_writer.writerow(compliance_row.__dict__)



from typing import Optional

from pydantic import BaseModel, validator

from prowler.config.config import prowler_version


class ProductFields(BaseModel):
    ProviderName: str = "Prowler"
    ProviderVersion: str = prowler_version
    ProwlerResourceName: str


class Severity(BaseModel):
    Label: str


class Resource(BaseModel):
    Type: str
    Id: str
    Partition: str
    Region: str
    Tags: Optional[dict]


class Compliance(BaseModel):
    Status: str
    RelatedRequirements: list[str]
    AssociatedStandards: list[dict]


class Recommendation(BaseModel):
    Text: str = ""
    Url: str = ""

    @validator("Text", pre=True, always=True)
    def text_must_not_exceed_512_chars(text):
        text_validated = text
        if len(text) > 512:
            text_validated = text[:509] + "..."
        return text_validated

    @validator("Url", pre=True, always=True)
    def set_default_url_if_empty(url):
        default_url = "https://docs.aws.amazon.com/securityhub/latest/userguide/what-is-securityhub.html"
        if url:
            default_url = url
        return default_url


class Remediation(BaseModel):
    Recommendation: Recommendation


class Check_Output_JSON_ASFF(BaseModel):
    """
    Check_Output_JSON_ASFF generates a finding's output in JSON ASFF format: https://docs.aws.amazon.com/securityhub/latest/userguide/securityhub-findings-format-syntax.html
    """

    SchemaVersion: str = "2018-10-08"
    Id: str
    ProductArn: str
    RecordState: str = "ACTIVE"
    ProductFields: ProductFields
    GeneratorId: str
    AwsAccountId: str
    Types: list[str] = None
    FirstObservedAt: str
    UpdatedAt: str
    CreatedAt: str
    Severity: Severity
    Title: str
    Description: str
    Resources: list[Resource] = None
    Compliance: Compliance
    Remediation: Remediation

    @validator("Description", pre=True, always=True)
    def description_must_not_exceed_1024_chars(description):
        description_validated = description
        if len(description) > 1024:
            description_validated = description[:1021] + "..."
        return description_validated



import os
import re
import sys

from colorama import Fore, Style
from google import auth
from google.oauth2.credentials import Credentials
from googleapiclient import discovery
from googleapiclient.errors import HttpError

from prowler.config.config import load_and_validate_config_file
from prowler.lib.logger import logger
from prowler.lib.utils.utils import print_boxes
from prowler.providers.common.models import Audit_Metadata
from prowler.providers.common.provider import Provider
from prowler.providers.gcp.models import (
    GCPIdentityInfo,
    GCPOrganization,
    GCPOutputOptions,
    GCPProject,
)


class GcpProvider(Provider):
    _type: str = "gcp"
    _session: Credentials
    _project_ids: list
    _excluded_project_ids: list
    _identity: GCPIdentityInfo
    _audit_config: dict
    _output_options: GCPOutputOptions
    _mutelist: dict
    # TODO: this is not optional, enforce for all providers
    audit_metadata: Audit_Metadata

    def __init__(self, arguments):
        logger.info("Instantiating GCP Provider ...")
        input_project_ids = arguments.project_id
        excluded_project_ids = arguments.excluded_project_id
        credentials_file = arguments.credentials_file
        list_project_ids = arguments.list_project_id

        self._session = self.setup_session(credentials_file)

        self._project_ids = []
        self._projects = {}
        self._excluded_project_ids = []
        accessible_projects = self.get_projects()
        if not accessible_projects:
            logger.critical("No Project IDs can be accessed via Google Credentials.")
            sys.exit(1)

        if input_project_ids:
            for input_project in input_project_ids:
                for accessible_project in accessible_projects:
                    if self.is_project_matching(input_project, accessible_project):
                        self._projects[accessible_project] = accessible_projects[
                            accessible_project
                        ]
                        self._project_ids.append(
                            accessible_projects[accessible_project].id
                        )
        else:
            # If not projects were input, all accessible projects are scanned by default
            for project_id, project in accessible_projects.items():
                self._projects[project_id] = project
                self._project_ids.append(project_id)

        # Remove excluded projects if any input
        if excluded_project_ids:
            for excluded_project in excluded_project_ids:
                for project_id in self._project_ids:
                    if self.is_project_matching(excluded_project, project_id):
                        self._excluded_project_ids.append(project_id)
            for project_id in self._excluded_project_ids:
                self._projects.pop(project_id)
                self._project_ids.remove(project_id)

        if not self._projects:
            logger.critical(
                "No Input Project IDs can be accessed via Google Credentials."
            )
            sys.exit(1)

        if list_project_ids:
            print(
                f"{Fore.YELLOW}Available GCP Project IDs{Style.RESET_ALL}:\n{' '.join(self._project_ids)}\n"
            )
            sys.exit(0)

        # Update organizations info
        self.update_projects_with_organizations()

        self._identity = GCPIdentityInfo(
            profile=getattr(self.session, "_service_account_email", "default"),
        )

        # TODO: move this to the providers, pending for AWS, GCP, AZURE and K8s
        # Audit Config
        self._audit_config = load_and_validate_config_file(
            self._type, arguments.config_file
        )
        self._fixer_config = load_and_validate_config_file(
            self._type, arguments.fixer_config
        )

    @property
    def identity(self):
        return self._identity

    @property
    def type(self):
        return self._type

    @property
    def session(self):
        return self._session

    @property
    def projects(self):
        return self._projects

    @property
    def project_ids(self):
        return self._project_ids

    @property
    def excluded_project_ids(self):
        return self._excluded_project_ids

    @property
    def audit_config(self):
        return self._audit_config

    @property
    def fixer_config(self):
        return self._fixer_config

    @property
    def output_options(self):
        return self._output_options

    @output_options.setter
    def output_options(self, options: tuple):
        arguments, bulk_checks_metadata = options
        self._output_options = GCPOutputOptions(
            arguments, bulk_checks_metadata, self._identity
        )

    @property
    def get_output_mapping(self):
        return {
            # Account: identity.profile
            "auth_method": "identity.profile",
            "provider": "type",
            # TODO: comes from finding, finding.project_id
            # "account_uid": "",
            # TODO: get project name from GCP
            # "account_name": "organizations_metadata.account_details_name",
            # There is no concept as project email in GCP
            # "account_email": "organizations_metadata.account_details_email",
            # TODO: get project organization ID from GCP
            # "account_organization_uid": "organizations_metadata.account_details_arn",
            # TODO: get project organization from GCP
            # "account_organization": "",
            # TODO: get project tags organization from GCP
            # "account_tags": "organizations_metadata.account_details_tags",
            # "partition": "identity.partition",
        }

    def setup_session(self, credentials_file):
        try:
            if credentials_file:
                self.__set_gcp_creds_env_var__(credentials_file)

            credentials, _ = auth.default(
                scopes=["https://www.googleapis.com/auth/cloud-platform"]
            )
            return credentials
        except Exception as error:
            logger.critical(
                f"{error.__class__.__name__}[{error.__traceback__.tb_lineno}]: {error}"
            )
            sys.exit(1)

    def __set_gcp_creds_env_var__(self, credentials_file):
        logger.info(
            "GCP provider: Setting GOOGLE_APPLICATION_CREDENTIALS environment variable..."
        )
        client_secrets_path = os.path.abspath(credentials_file)
        os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = client_secrets_path

    def print_credentials(self):
        # TODO: Beautify audited profile, set "default" if there is no profile set
        # TODO: improve print_credentials with more data like name, number, organization
        report_lines = [
            f"GCP Account: {Fore.YELLOW}{self.identity.profile}{Style.RESET_ALL}",
            f"GCP Project IDs: {Fore.YELLOW}{', '.join(self.project_ids)}{Style.RESET_ALL}",
        ]
        if self.excluded_project_ids:
            report_lines.append(
                f"Excluded GCP Project IDs: {Fore.YELLOW}{', '.join(self.excluded_project_ids)}{Style.RESET_ALL}"
            )
        report_title = (
            f"{Style.BRIGHT}Using the GCP credentials below:{Style.RESET_ALL}"
        )
        print_boxes(report_lines, report_title)

    def get_projects(self) -> dict[str, GCPProject]:
        try:
            projects = {}

            service = discovery.build(
                "cloudresourcemanager", "v1", credentials=self.session
            )

            request = service.projects().list()

            while request is not None:
                response = request.execute()

                for project in response.get("projects", []):
                    labels = []
                    for key, value in project.get("labels", {}).items():
                        labels.append(f"{key}:{value}")

                    project_id = project["projectId"]
                    gcp_project = GCPProject(
                        number=project["projectNumber"],
                        id=project_id,
                        name=project["name"],
                        lifecycle_state=project["lifecycleState"],
                        labels=labels,
                    )

                    if (
                        "parent" in project
                        and "type" in project["parent"]
                        and project["parent"]["type"] == "organization"
                    ):
                        organization_id = project["parent"]["id"]
                        gcp_project.organization = GCPOrganization(
                            id=organization_id, name=f"organizations/{organization_id}"
                        )

                    projects[project_id] = gcp_project
                request = service.projects().list_next(
                    previous_request=request, previous_response=response
                )

        except HttpError as http_error:
            if "Cloud Resource Manager API has not been used" in str(http_error):
                logger.critical(
                    "Cloud Resource Manager API has not been used before or it is disabled. Enable it by visiting https://console.developers.google.com/apis/api/cloudresourcemanager.googleapis.com/ then retry."
                )
                sys.exit(1)
            else:
                logger.error(
                    f"{http_error.__class__.__name__}[{http_error.__traceback__.tb_lineno}]: {http_error}"
                )
        except Exception as error:
            if error.__class__.__name__ == "RefreshError":
                logger.critical(
                    "Google Cloud SDK has not been authenticated or the credentials have expired. Authenticate by running 'gcloud auth application-default login' then retry."
                )
            else:
                logger.critical(
                    f"{error.__class__.__name__}[{error.__traceback__.tb_lineno}]: {error}"
                )
            sys.exit(1)
        finally:
            return projects

    def update_projects_with_organizations(self):
        try:
            service = discovery.build(
                "cloudresourcemanager", "v1", credentials=self._session
            )
            # TODO: this call requires more permissions to get that data
            # resourcemanager.organizations.get --> add to the docs
            for project in self._projects.values():
                if project.organization:
                    request = service.organizations().get(
                        name=f"organizations/{project.organization.id}"
                    )

                    while request is not None:
                        response = request.execute()
                        project.organization.display_name = response.get("displayName")
                        request = service.projects().list_next(
                            previous_request=request, previous_response=response
                        )

        except HttpError as http_error:
            if http_error.status_code == 403 and "organizations" in http_error.uri:
                logger.error(
                    f"{http_error.__class__.__name__}[{http_error.__traceback__.tb_lineno}]: {http_error.error_details} to get Organizations display name."
                )
            else:
                logger.error(
                    f"{http_error.__class__.__name__}[{http_error.__traceback__.tb_lineno}]: {http_error}"
                )
        except Exception as error:
            logger.error(
                f"{error.__class__.__name__}[{error.__traceback__.tb_lineno}]: {error}"
            )

    def is_project_matching(self, input_project: str, project_to_match: str) -> bool:
        """
        Check if the input project matches the project to match
        Args:
            input_project: str
            project_to_match: str
        Returns:
            bool
        """
        return (
            "*" in input_project
            and re.search(
                "." + input_project if input_project.startswith("*") else input_project,
                project_to_match,
            )
        ) or input_project == project_to_match



def is_policy_cross_account(policy: dict, audited_account: str) -> bool:
    """
    is_policy_cross_account checks if the policy allows cross-account access.
    Args:
        policy (dict): The policy to check.
        audited_account (str): The account to check if it has access.
    Returns:
        bool: True if the policy allows cross-account access, False otherwise.
    """
    if policy and "Statement" in policy:
        if isinstance(policy["Statement"], list):
            for statement in policy["Statement"]:
                if statement["Effect"] == "Allow" and "AWS" in statement["Principal"]:
                    if isinstance(statement["Principal"]["AWS"], list):
                        for aws_account in statement["Principal"]["AWS"]:
                            if audited_account not in aws_account or "*" == aws_account:
                                return True
                    else:
                        if (
                            audited_account not in statement["Principal"]["AWS"]
                            or "*" == statement["Principal"]["AWS"]
                        ):
                            return True
        else:
            statement = policy["Statement"]
            if statement["Effect"] == "Allow" and "AWS" in statement["Principal"]:
                if isinstance(statement["Principal"]["AWS"], list):
                    for aws_account in statement["Principal"]["AWS"]:
                        if audited_account not in aws_account or "*" == aws_account:
                            return True
                else:
                    if (
                        audited_account not in statement["Principal"]["AWS"]
                        or "*" == statement["Principal"]["AWS"]
                    ):
                        return True
    return False


def is_policy_public(policy: dict) -> bool:
    """
    is_policy_public checks if the policy is publicly accessible.
    If the "Principal" element value is set to { "AWS": "*" } and the policy statement is not using any Condition clauses to filter the access, the selected policy is publicly accessible.
    Args:
        policy (dict): The policy to check.
        Returns:
        bool: True if the policy is publicly accessible, False otherwise.
    """
    if policy and "Statement" in policy:
        for statement in policy["Statement"]:
            if (
                "Principal" in statement
                and (
                    "*" == statement["Principal"]
                    or "arn:aws:iam::*:root" in statement["Principal"]
                )
                and "Condition" not in statement
            ):
                return True
            elif "Principal" in statement and "AWS" in statement["Principal"]:
                if isinstance(statement["Principal"]["AWS"], str):
                    principals = [statement["Principal"]["AWS"]]
                else:
                    principals = statement["Principal"]["AWS"]
                for principal_arn in principals:
                    if (
                        principal_arn == "*" or principal_arn == "arn:aws:iam::*:root"
                    ) and "Condition" not in statement:
                        return True
    return False



from prowler.lib.check.models import Check, Check_Report_AWS
from prowler.providers.aws.services.cloudtrail.cloudtrail_client import (
    cloudtrail_client,
)


class cloudtrail_multi_region_enabled_logging_management_events(Check):
    def execute(self):
        findings = []
        if cloudtrail_client.trails is not None:
            report = Check_Report_AWS(self.metadata())
            report.status = "FAIL"
            report.status_extended = "No trail found with multi-region enabled and logging management events."
            report.region = cloudtrail_client.region
            report.resource_id = cloudtrail_client.audited_account
            report.resource_arn = cloudtrail_client.trail_arn_template

            for trail in cloudtrail_client.trails.values():
                if trail.is_logging:
                    if trail.is_multiregion:
                        for event in trail.data_events:
                            # Classic event selectors
                            if not event.is_advanced:
                                # Check if trail has IncludeManagementEvents and ReadWriteType is All
                                if (
                                    event.event_selector["ReadWriteType"] == "All"
                                    and event.event_selector["IncludeManagementEvents"]
                                ):
                                    report.region = trail.region
                                    report.resource_id = trail.name
                                    report.resource_arn = trail.arn
                                    report.resource_tags = trail.tags
                                    report.status = "PASS"
                                    report.status_extended = f"Trail {trail.name} from home region {trail.home_region} is multi-region, is logging and have management events enabled."

                            # Advanced event selectors
                            elif event.is_advanced:
                                if event.event_selector.get(
                                    "Name"
                                ) == "Management events selector" and all(
                                    [
                                        field["Field"] != "readOnly"
                                        for field in event.event_selector[
                                            "FieldSelectors"
                                        ]
                                    ]
                                ):
                                    report.region = trail.region
                                    report.resource_id = trail.name
                                    report.resource_arn = trail.arn
                                    report.resource_tags = trail.tags
                                    report.status = "PASS"
                                    report.status_extended = f"Trail {trail.name} from home region {trail.home_region} is multi-region, is logging and have management events enabled."
            findings.append(report)
        return findings



from csv import DictWriter

from prowler.config.config import timestamp
from prowler.lib.logger import logger
from prowler.lib.outputs.compliance.models import Check_Output_CSV_AWS_Well_Architected
from prowler.lib.outputs.csv.csv import generate_csv_fields
from prowler.lib.utils.utils import outputs_unix_timestamp


def write_compliance_row_aws_well_architected_framework(
    file_descriptors, finding, compliance, output_options, provider
):
    try:
        compliance_output = compliance.Framework
        if compliance.Version != "":
            compliance_output += "_" + compliance.Version
        if compliance.Provider != "":
            compliance_output += "_" + compliance.Provider
        compliance_output = compliance_output.lower().replace("-", "_")
        csv_header = generate_csv_fields(Check_Output_CSV_AWS_Well_Architected)
        csv_writer = DictWriter(
            file_descriptors[compliance_output],
            fieldnames=csv_header,
            delimiter=";",
        )
        for requirement in compliance.Requirements:
            requirement_description = requirement.Description
            requirement_id = requirement.Id
            for attribute in requirement.Attributes:
                compliance_row = Check_Output_CSV_AWS_Well_Architected(
                    Provider=finding.check_metadata.Provider,
                    Description=compliance.Description,
                    AccountId=provider.identity.account,
                    Region=finding.region,
                    AssessmentDate=outputs_unix_timestamp(
                        output_options.unix_timestamp, timestamp
                    ),
                    Requirements_Id=requirement_id,
                    Requirements_Description=requirement_description,
                    Requirements_Attributes_Name=attribute.Name,
                    Requirements_Attributes_WellArchitectedQuestionId=attribute.WellArchitectedQuestionId,
                    Requirements_Attributes_WellArchitectedPracticeId=attribute.WellArchitectedPracticeId,
                    Requirements_Attributes_Section=attribute.Section,
                    Requirements_Attributes_SubSection=attribute.SubSection,
                    Requirements_Attributes_LevelOfRisk=attribute.LevelOfRisk,
                    Requirements_Attributes_AssessmentMethod=attribute.AssessmentMethod,
                    Requirements_Attributes_Description=attribute.Description,
                    Requirements_Attributes_ImplementationGuidanceUrl=attribute.ImplementationGuidanceUrl,
                    Status=finding.status,
                    StatusExtended=finding.status_extended,
                    ResourceId=finding.resource_id,
                    CheckId=finding.check_metadata.CheckID,
                    Muted=finding.muted,
                )

                csv_writer.writerow(compliance_row.__dict__)
    except Exception as error:
        logger.error(
            f"{error.__class__.__name__}[{error.__traceback__.tb_lineno}]: {error}"
        )



from csv import DictWriter

from colorama import Fore, Style
from tabulate import tabulate

from prowler.config.config import orange_color, timestamp
from prowler.lib.outputs.compliance.models import Check_Output_CSV_Generic_Compliance
from prowler.lib.outputs.csv.csv import generate_csv_fields
from prowler.lib.utils.utils import outputs_unix_timestamp


def write_compliance_row_generic(
    file_descriptors, finding, compliance, output_options, provider
):
    compliance_output = compliance.Framework
    if compliance.Version != "":
        compliance_output += "_" + compliance.Version
    if compliance.Provider != "":
        compliance_output += "_" + compliance.Provider

    compliance_output = compliance_output.lower().replace("-", "_")
    csv_header = generate_csv_fields(Check_Output_CSV_Generic_Compliance)
    csv_writer = DictWriter(
        file_descriptors[compliance_output],
        fieldnames=csv_header,
        delimiter=";",
    )
    for requirement in compliance.Requirements:
        requirement_description = requirement.Description
        requirement_id = requirement.Id
        for attribute in requirement.Attributes:
            compliance_row = Check_Output_CSV_Generic_Compliance(
                Provider=finding.check_metadata.Provider,
                Description=compliance.Description,
                AccountId=provider.identity.account,
                Region=finding.region,
                AssessmentDate=outputs_unix_timestamp(
                    output_options.unix_timestamp, timestamp
                ),
                Requirements_Id=requirement_id,
                Requirements_Description=requirement_description,
                Requirements_Attributes_Section=attribute.Section,
                Requirements_Attributes_SubSection=attribute.SubSection,
                Requirements_Attributes_SubGroup=attribute.SubGroup,
                Requirements_Attributes_Service=attribute.Service,
                Requirements_Attributes_Type=attribute.Type,
                Status=finding.status,
                StatusExtended=finding.status_extended,
                ResourceId=finding.resource_id,
                CheckId=finding.check_metadata.CheckID,
                Muted=finding.muted,
            )
            csv_writer.writerow(compliance_row.__dict__)


def get_generic_compliance_table(
    findings: list,
    bulk_checks_metadata: dict,
    compliance_framework: str,
    output_filename: str,
    output_directory: str,
    compliance_overview: bool,
):
    pass_count = []
    fail_count = []
    muted_count = []
    for index, finding in enumerate(findings):
        check = bulk_checks_metadata[finding.check_metadata.CheckID]
        check_compliances = check.Compliance
        for compliance in check_compliances:
            if (
                compliance.Framework.upper()
                in compliance_framework.upper().replace("_", "-")
                and compliance.Version in compliance_framework.upper()
                and compliance.Provider in compliance_framework.upper()
            ):
                for requirement in compliance.Requirements:
                    for attribute in requirement.Attributes:
                        if finding.muted:
                            if index not in muted_count:
                                muted_count.append(index)
                        else:
                            if finding.status == "FAIL" and index not in fail_count:
                                fail_count.append(index)
                            elif finding.status == "PASS" and index not in pass_count:
                                pass_count.append(index)
    if (
        len(fail_count) + len(pass_count) + len(muted_count) > 1
    ):  # If there are no resources, don't print the compliance table
        print(
            f"\nCompliance Status of {Fore.YELLOW}{compliance_framework.upper()}{Style.RESET_ALL} Framework:"
        )
        overview_table = [
            [
                f"{Fore.RED}{round(len(fail_count) / len(findings) * 100, 2)}% ({len(fail_count)}) FAIL{Style.RESET_ALL}",
                f"{Fore.GREEN}{round(len(pass_count) / len(findings) * 100, 2)}% ({len(pass_count)}) PASS{Style.RESET_ALL}",
                f"{orange_color}{round(len(muted_count) / len(findings) * 100, 2)}% ({len(muted_count)}) MUTED{Style.RESET_ALL}",
            ]
        ]
        print(tabulate(overview_table, tablefmt="rounded_grid"))
        if not compliance_overview:
            print(f"\nDetailed results of {compliance_framework.upper()} are in:")
            print(
                f" - CSV: {output_directory}/compliance/{output_filename}_{compliance_framework}.csv\n"
            )



from csv import DictWriter
from importlib import import_module

from colorama import Fore, Style
from tabulate import tabulate

from prowler.config.config import orange_color, timestamp
from prowler.lib.logger import logger
from prowler.lib.outputs.csv.csv import generate_csv_fields
from prowler.lib.outputs.utils import unroll_list
from prowler.lib.utils.utils import outputs_unix_timestamp


def write_compliance_row_mitre_attack(file_descriptors, finding, compliance, provider):
    try:
        compliance_output = compliance.Framework
        if compliance.Version != "":
            compliance_output += "_" + compliance.Version
        if compliance.Provider != "":
            compliance_output += "_" + compliance.Provider

        mitre_attack_model_name = "MitreAttack" + compliance.Provider
        module = import_module("prowler.lib.outputs.compliance.mitre_attack.models")
        mitre_attack_model = getattr(module, mitre_attack_model_name)
        compliance_output = compliance_output.lower().replace("-", "_")
        csv_header = generate_csv_fields(mitre_attack_model)
        csv_writer = DictWriter(
            file_descriptors[compliance_output],
            fieldnames=csv_header,
            delimiter=";",
        )
        for requirement in compliance.Requirements:

            if compliance.Provider == "AWS":
                attributes_services = ", ".join(
                    attribute.AWSService for attribute in requirement.Attributes
                )
            elif compliance.Provider == "Azure":
                attributes_services = ", ".join(
                    attribute.AzureService for attribute in requirement.Attributes
                )
            elif compliance.Provider == "GCP":
                attributes_services = ", ".join(
                    attribute.GCPService for attribute in requirement.Attributes
                )
            requirement_description = requirement.Description
            requirement_id = requirement.Id
            requirement_name = requirement.Name
            attributes_categories = ", ".join(
                attribute.Category for attribute in requirement.Attributes
            )
            attributes_values = ", ".join(
                attribute.Value for attribute in requirement.Attributes
            )
            attributes_comments = ", ".join(
                attribute.Comment for attribute in requirement.Attributes
            )

            common_data = {
                "Provider": finding.check_metadata.Provider,
                "Description": compliance.Description,
                "AssessmentDate": outputs_unix_timestamp(
                    provider.output_options.unix_timestamp, timestamp
                ),
                "Requirements_Id": requirement_id,
                "Requirements_Name": requirement_name,
                "Requirements_Description": requirement_description,
                "Requirements_Tactics": unroll_list(requirement.Tactics),
                "Requirements_SubTechniques": unroll_list(requirement.SubTechniques),
                "Requirements_Platforms": unroll_list(requirement.Platforms),
                "Requirements_TechniqueURL": requirement.TechniqueURL,
                "Requirements_Attributes_Services": attributes_services,
                "Requirements_Attributes_Categories": attributes_categories,
                "Requirements_Attributes_Values": attributes_values,
                "Requirements_Attributes_Comments": attributes_comments,
                "Status": finding.status,
                "StatusExtended": finding.status_extended,
                "ResourceId": finding.resource_id,
                "CheckId": finding.check_metadata.CheckID,
                "Muted": finding.muted,
            }
            if compliance.Provider == "AWS":
                common_data["AccountId"] = provider.identity.account
                common_data["Region"] = finding.region
            elif compliance.Provider == "Azure":
                common_data["SubscriptionId"] = unroll_list(
                    provider.identity.subscriptions
                )
            elif compliance.Provider == "GCP":
                common_data["ProjectId"] = unroll_list(provider.projects)

            compliance_row = mitre_attack_model(**common_data)

            csv_writer.writerow(compliance_row.__dict__)
    except Exception as error:
        logger.error(
            f"{error.__class__.__name__}[{error.__traceback__.tb_lineno}]: {error}"
        )


def get_mitre_attack_table(
    findings: list,
    bulk_checks_metadata: dict,
    compliance_framework: str,
    output_filename: str,
    output_directory: str,
    compliance_overview: bool,
):
    tactics = {}
    mitre_compliance_table = {
        "Provider": [],
        "Tactic": [],
        "Status": [],
        "Muted": [],
    }
    pass_count = []
    fail_count = []
    muted_count = []
    for index, finding in enumerate(findings):
        check = bulk_checks_metadata[finding.check_metadata.CheckID]
        check_compliances = check.Compliance
        for compliance in check_compliances:
            if (
                "MITRE-ATTACK" in compliance.Framework
                and compliance.Version in compliance_framework
            ):
                for requirement in compliance.Requirements:
                    for tactic in requirement.Tactics:
                        if tactic not in tactics:
                            tactics[tactic] = {"FAIL": 0, "PASS": 0, "Muted": 0}
                        if finding.muted:
                            if index not in muted_count:
                                muted_count.append(index)
                                tactics[tactic]["Muted"] += 1
                        else:
                            if finding.status == "FAIL":
                                if index not in fail_count:
                                    fail_count.append(index)
                                    tactics[tactic]["FAIL"] += 1
                            elif finding.status == "PASS":
                                if index not in pass_count:
                                    pass_count.append(index)
                                    tactics[tactic]["PASS"] += 1
    # Add results to table
    tactics = dict(sorted(tactics.items()))
    for tactic in tactics:
        mitre_compliance_table["Provider"].append(compliance.Provider)
        mitre_compliance_table["Tactic"].append(tactic)
        if tactics[tactic]["FAIL"] > 0:
            mitre_compliance_table["Status"].append(
                f"{Fore.RED}FAIL({tactics[tactic]['FAIL']}){Style.RESET_ALL}"
            )
        else:
            mitre_compliance_table["Status"].append(
                f"{Fore.GREEN}PASS({tactics[tactic]['PASS']}){Style.RESET_ALL}"
            )
        mitre_compliance_table["Muted"].append(
            f"{orange_color}{tactics[tactic]['Muted']}{Style.RESET_ALL}"
        )
    if (
        len(fail_count) + len(pass_count) + len(muted_count) > 1
    ):  # If there are no resources, don't print the compliance table
        print(
            f"\nCompliance Status of {Fore.YELLOW}{compliance_framework.upper()}{Style.RESET_ALL} Framework:"
        )
        overview_table = [
            [
                f"{Fore.RED}{round(len(fail_count) / len(findings) * 100, 2)}% ({len(fail_count)}) FAIL{Style.RESET_ALL}",
                f"{Fore.GREEN}{round(len(pass_count) / len(findings) * 100, 2)}% ({len(pass_count)}) PASS{Style.RESET_ALL}",
                f"{orange_color}{round(len(muted_count) / len(findings) * 100, 2)}% ({len(muted_count)}) MUTED{Style.RESET_ALL}",
            ]
        ]
        print(tabulate(overview_table, tablefmt="rounded_grid"))
        if not compliance_overview:
            print(
                f"\nFramework {Fore.YELLOW}{compliance_framework.upper()}{Style.RESET_ALL} Results:"
            )
            print(
                tabulate(
                    mitre_compliance_table,
                    headers="keys",
                    tablefmt="rounded_grid",
                )
            )
            print(
                f"{Style.BRIGHT}* Only sections containing results appear.{Style.RESET_ALL}"
            )
            print(f"\nDetailed results of {compliance_framework.upper()} are in:")
            print(
                f" - CSV: {output_directory}/compliance/{output_filename}_{compliance_framework}.csv\n"
            )



import os
import sys

from prowler.config.config import (
    json_asff_file_suffix,
    json_file_suffix,
    json_ocsf_file_suffix,
)
from prowler.lib.logger import logger
from prowler.lib.utils.utils import open_file


def close_json(output_filename, output_directory, mode):
    """close_json closes the output JSON file replacing the last comma with ]"""
    try:
        suffix = json_file_suffix
        if mode == "json-asff":
            suffix = json_asff_file_suffix
        elif mode == "json-ocsf":
            suffix = json_ocsf_file_suffix
        filename = f"{output_directory}/{output_filename}{suffix}"
        # Close JSON file if exists
        if os.path.isfile(filename):
            file_descriptor = open_file(
                filename,
                "a",
            )
            # Replace last comma for square bracket if not empty
            if file_descriptor.tell() > 0:
                if file_descriptor.tell() != 1:
                    file_descriptor.seek(file_descriptor.tell() - 1, os.SEEK_SET)
                file_descriptor.truncate()
                file_descriptor.write("]")
            file_descriptor.close()
    except Exception as error:
        logger.critical(
            f"{error.__class__.__name__}[{error.__traceback__.tb_lineno}] -- {error}"
        )
        sys.exit(1)
